package org.colllib.builder;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Vector;

import org.colllib.util.CollectionUtil;


/**
 * A little helper class to create filled lists on the fly. Create a given
 * builder via <code>create</code> or one of its standard variants, call the
 * various add methods, very much like a string builder, retrieve the finished
 * list via <code>build()</code>
 * <p>
 * <pre>
 *    System.out.println(
 *      ListBuilder.&lt;String&gt;createArray()
 *         .add("Hello")
 *         .add(" ", "World")
 *         .build());
 *    ==> { "Hello", " ", "World" }
 * </pre>
 * <p>
 * This file is part of CollLib released under the terms of the LGPL V3.0.
 * See the file licenses/lgpl-3.0.txt for details.
 *
 * @author marc.jackisch
 */
public class ListBuilder<E, L extends List<E>> {
   /**
    * Create a list builder for the given list
    * 
    * @param <E>  element type
    * @param <L>  list type
    * @param l    the list
    * @return a list builder that operates on the list l
    */
   public static <E, L extends List<E>> ListBuilder<E, L> create(L l) {
      return new ListBuilder<E, L>(l);
   }
   
   /**
    * Create a list builder for a new array list
    * 
    * @param <E>  element type
    * @return a list builder for an ArrayList&lt;E&gt;
    */
   public static <E> ListBuilder<E, ArrayList<E>> createArray() {
      return create(new ArrayList<E>());
   }

   /**
    * Create a list builder for a new linked list
    * 
    * @param <E>  element type
    * @return a list builder for a LinkedList&lt;E&gt;
    */
   public static <E> ListBuilder<E, LinkedList<E>> createLinked() {
      return create(new LinkedList<E>());
   }

   /**
    * Create a list builder for a new vector
    * 
    * @param <E>  element type
    * @return a list builder for a Vector&lt;E&gt;
    */
   public static <E> ListBuilder<E, Vector<E>> createVector() {
      return create(new Vector<E>());
   }

   private final L l;

   private ListBuilder(L list) {
      l = list;
   }

   /**
    * Add an element to the underlying list
    * 
    * @param el  the element
    * @return this
    */
   public ListBuilder<E, L> add(E el) {
      l.add(el);
      return this;
   }

   /**
    * Add elements to the underlying list
    * 
    * @param els the elements
    * @return this
    */
   public ListBuilder<E, L> add(E ... els) {
      CollectionUtil.addAll(l, els);
      return this;
   }

   /**
    * Add elements to the underlying list
    * 
    * @param els the elements
    * @return this
    */
   public ListBuilder<E, L> add(Iterable<E> els) {
      CollectionUtil.addAll(l, els);
      return this;
   }

   /**
    * Fetch the underlying list
    * 
    * @return the underlying list
    */
   public L build() {
      return l;
   }
}
