package org.colllib.builder;

import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.Set;
import java.util.TreeSet;

import org.colllib.util.CollectionUtil;

/**
 * A class to create filled sets on the fly. Set up a builder by calling
 * the <code>create</code> method with a given set or choose a standard 
 * variant by callig the appropriate constructor.<br>
 * Call the different <code>add</code> methods and get the constructed 
 * set via <code>build()</code>.
 * <p>
 * <pre>
 *    System.out.println(
 *      ListBuilder.&lt;String&gt;createHash()
 *         .add("a")
 *         .add("b", "c")
 *         .build());
 *    ==> { "b", "c", "a" }
 * </pre>
 * <p>
 * This file is part of CollLib released under the terms of the LGPL V3.0.
 * See the file licenses/lgpl-3.0.txt for details.
 * 
 * @author marcel.billen
 */
public class SetBuilder<E, S extends Set<E>> {
   /**
    * Create a set builder for the given set
    * 
    * @param <E>  element type
    * @param <S>  set type
    * @param s    the set
    * @return a set builder that operates on the set s
    */
   public static <E, S extends Set<E>> SetBuilder<E, S> create(S s) {
      return new SetBuilder<E, S>(s);
   }
   
   /**
    * Create a set builder for a new hash set
    * 
    * @param <E>  element type
    * @return a set builder for a HashSet&lt;E&gt;
    */
   public static <E> SetBuilder<E, HashSet<E>> createHash() {
      return create(new HashSet<E>());
   }
   
   /**
    * Create a set builder for a new linked hash set
    * 
    * @param <E>  element type
    * @return a set builder for a LinkedHashSet&lt;E&gt;
    */
   public static <E> SetBuilder<E, LinkedHashSet<E>> createLinkedHash() {
      return create(new LinkedHashSet<E>());
   }
   
   /**
    * Create a set builder for a new tree set
    * 
    * @param <E>  element type
    * @return a set builder for a TreeSet&lt;E&gt;
    */
   public static <E> SetBuilder<E, TreeSet<E>> createTree() {
      return create(new TreeSet<E>());
   }
   
   private final S s;
   
   private SetBuilder(S set) {
      s = set;
   }
   
   /**
    * Add an element to the underlying set
    * 
    * @param el  the element
    * @return this
    */
   public SetBuilder<E, S> add(E el) {
      s.add(el);
      return this;
   }

   /**
    * Add elements to the underlying set
    * 
    * @param els the elements
    * @return this
    */
   public SetBuilder<E, S> add(E ... els) {
      CollectionUtil.addAll(s, els);
      return this;
   }
   
   /**
    * Add elements to the underlying set
    * 
    * @param els the elements
    * @return this
    */
   public SetBuilder<E, S> add(Iterable<E> els) {
      CollectionUtil.addAll(s, els);
      return this;
   }
   
   /**
    * Fetch the underlying set
    * 
    * @return the underlying set
    */
   public S build() {
      return s;
   }
}

