package org.colllib.caches;

import java.util.HashMap;

/**
 * A generic cache, using a lookup provider implementation to fetch the values
 * of single objects.
 * <p>
 * This file is part of CollLib released under the terms of the LGPL V3.0.
 * See the file licenses/lgpl-3.0.txt for details.
 *
 * @author marc.jackisch
 */
public class GenCache<K, V> {
   private LookupProvider<K, V> lookup;
   private HashMap<K, V>        map;

   /**
    * Create a new GenCache using the specified lookup provider
    * 
    * @param lookup  the lookup provider implementation
    */
   public GenCache(LookupProvider<K, V> lookup) {
      this.lookup = lookup;
      map = new HashMap<K, V>();
   }

   /**
    * Fetch a value from the cache. Will use the cached value, if the key has
    * already been queried, otherwise will branch to the lookup provider and
    * return its result (storing it for future queries)
    * 
    * @param k   the key
    * @return    the value for k
    */
   public V get(K k) {
      if(map.containsKey(k))
         return map.get(k);
      V v = lookup.lookup(k);
      map.put(k, v);
      return v;
   }

   /**
    * Check, whether the key k is already contained (i.e. has been queried)
    * in this cache.
    * 
    * @param k  the key
    * @return   map.containsKey
    */
   public boolean containsKey(K k) {
      return map.containsKey(k);
   }

   /**
    * Clear the cache. All keys will be queried from the lookup provider once
    * more.
    */
   public void clear() {
      map.clear();
   }

   /**
    * Lookup provider interface
    */
   public static interface LookupProvider<K, V> {
      /**
       * Fetch a key from whatever backing implementation
       * @param k   the key
       * @return    the value for k
       */
      public V lookup(K k);
   };
}
