package org.colllib.datastruct;

import java.io.Serializable;
import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.NoSuchElementException;

/**
 * A rudimentary list-wrapper over a given backing array, which implements an
 * immutable list on the given array without copying it.
 * <p>
 * This file is part of CollLib released under the terms of the LGPL V3.0.
 * See the file licenses/lgpl-3.0.txt for details.
 * 
 * @author marc.jackisch
 */
public class ImmutableArrayBackedList<E> implements List<E>, Serializable {
   private E[] a;

   public ImmutableArrayBackedList(E[] array) {
      this.a = array;
   }

   @Override
   public int size() {
      return a.length;
   }

   @Override
   public boolean isEmpty() {
      return a.length == 0;
   }

   @Override
   public boolean contains(Object o) {
      return indexOf(o) >= 0;
   }

   @Override
   public Iterator<E> iterator() {
      return new Iterator<E>() {
         int i = 0;
         @Override
         public boolean hasNext() {
            return i < a.length;
         }

         @Override
         public E next() {
            try {
               return a[i++];
            }
            catch(IndexOutOfBoundsException e) {
               throw new NoSuchElementException();
            }
         }

         @Override
         public void remove() {
            throw new UnsupportedOperationException("remove");
         }};
   }

   @Override
   public Object[] toArray() {
      return a;
   }

   @SuppressWarnings("unchecked")
   @Override
   public <T> T[] toArray(T[] targetArray) {
      if(targetArray.length < a.length)
         return (T[]) Arrays.copyOf(a, a.length, a.getClass());
     System.arraycopy(a, 0, targetArray, 0, a.length);
     if(targetArray.length > a.length)
        targetArray[a.length] = null;
     return targetArray;
   }

   @Override
   public boolean add(E e) {
      throw new UnsupportedOperationException("add");
   }

   @Override
   public boolean remove(Object o) {
      throw new UnsupportedOperationException("remove");
   }

   @Override
   public boolean containsAll(Collection<?> c) {
      for(Object o : c)
         if(!contains(o))
            return false;
      return true;
   }

   @Override
   public boolean addAll(Collection<? extends E> c) {
      throw new UnsupportedOperationException("addAll");
   }

   @Override
   public boolean addAll(int index, Collection<? extends E> c) {
      throw new UnsupportedOperationException("addAll");
   }

   @Override
   public boolean removeAll(Collection<?> c) {
      throw new UnsupportedOperationException("removeAll");
   }

   @Override
   public boolean retainAll(Collection<?> c) {
      throw new UnsupportedOperationException("retainAll");
   }

   @Override
   public void clear() {
      throw new UnsupportedOperationException("clear");
   }

   @Override
   public E get(int index) {
      return a[index];
   }

   @Override
   public E set(int index, E element) {
      throw new UnsupportedOperationException("set");
   }

   @Override
   public void add(int index, E element) {
      throw new UnsupportedOperationException("add");
   }

   @Override
   public E remove(int index) {
      throw new UnsupportedOperationException("remove");
   }

   @Override
   public int indexOf(Object o) {
      for(int i = 0; i < a.length; i++)
         if(o == null && a[i] == null || (o != null && a[i] != null && o.equals(a[i])))
            return i;
      return -1;
   }

   @Override
   public int lastIndexOf(Object o) {
      for(int i = a.length - 1; i >= 0; i--)
         if(o == null && a[i] == null || (o != null && a[i] != null && o.equals(a[i])))
            return i;
      return -1;
   }

   @Override
   public ListIterator<E> listIterator() {
      throw new UnsupportedOperationException("listIterator");
   }

   @Override
   public ListIterator<E> listIterator(int index) {
      throw new UnsupportedOperationException("listIterator");
   }

   @Override
   public List<E> subList(int fromIndex, int toIndex) {
      @SuppressWarnings("unchecked")
      E[] s = (E[]) new Object[toIndex - fromIndex];
      System.arraycopy(a, fromIndex, s, 0, s.length);
      return new ImmutableArrayBackedList<E>(s);
   }
}
