package org.colllib.transformer;

import java.io.Serializable;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Map;

import org.apache.commons.beanutils.PropertyUtils;
import org.colllib.caches.ReadMethodCache;
import org.colllib.datastruct.Pair;
import org.colllib.util.ELAccessor;

/**
 * Various standard transformers.
 * 
 * This file is part of CollLib released under the terms of the LGPL V3.0.
 * See the file licenses/lgpl-3.0.txt for details.
 *
 * @author marc.jackisch
 */
public class TransformerCollection {
   private TransformerCollection() {}

   /**
    * Create a transformer, which transforms an object to an arbitrary property
    * value of that object. (Using reflection)
    *
    * @param <BeanClass>   the base object class
    * @param <ValueClass>  the returning class
    * @param propertyName  the property to read (must be readable via standard
    * bean methods, i.e. getXY or isXY)
    * @return a transformer
    */
   public static <BeanClass, ValueClass> Transformer<BeanClass, ValueClass> beanValue(String propertyName) {
      return new BeanValueTransformer<BeanClass, ValueClass>(propertyName);
   }

   private static class BeanValueTransformer<BeanClass, ValueClass> implements Transformer<BeanClass, ValueClass>, Serializable {
      private ReadMethodCache<ValueClass>  rmc;

      private BeanValueTransformer(String propertyName) {
         rmc = new ReadMethodCache<ValueClass>(propertyName);
      }

      @Override
      public ValueClass transform(BeanClass s) {
         if(s == null)
            return null;

         return rmc.invoke(s);
      }
   }

   /**
    * Create the identity transformer, which returns every input object unaltered
    * @param <T>  the base type
    * @return a does-nothing transformer
    */
   public static <T> Transformer<T, T> id() {
      return new IdTransformer<T>();
   }

   private static class IdTransformer<T> implements Transformer<T, T>, Serializable {
      private IdTransformer() {}

      @Override
      public T transform(T s) {
         return s;
      };
   }

   /**
    * Create a cast transformer, which casts the input object to the given
    * target class (and naturally throws a class cast exception, if it does
    * not match.)
    * 
    * @param <S>        the source class
    * @param <T>        the target class
    * @param tClass     the target class
    * @return a transformer
    */
   public static <S, T> Transformer<S, T> cast(Class<T> tClass) {
      return new CastTransformer<S, T>(tClass);
   }

   private static class CastTransformer<S, T> implements Transformer<S, T>, Serializable {
      private Class<T> tClass;

      public CastTransformer(Class<T> tClass) {
         this.tClass = tClass;
      }

      @Override
      public T transform(S s) {
         return s == null ? null : tClass.cast(s);
      }
   }

   /**
    * Transformer to invoke a one argument constructor of the destination class
    * with the source object
    *
    * @param <S>         the source class
    * @param <T>         the target class
    * @param sourceClass the source class
    * @param destClass   the target class
    * @return a transformer
    */
   public static <S, T> Transformer<S, T> constructorCall(Class<S> sourceClass, Class<T> destClass) {
      return new ConstructorCallTransformer<S, T>(sourceClass, destClass, new Class<?>[] { sourceClass }, 0, new Object[] { null });
   }

   /**
    * Transformer to invoke a one argument constructor of the destination class
    * with the source object
    *
    * @param <S>         the source class
    * @param <T>         the target class
    * @param sourceClass the source class
    * @param destClass   the target class
    * @param argTypes    the constructor argument types
    * @param index       the parameter index at which the transformation object is inserted
    * @param args        the (constant) constructor arguments
    * @return a transformer
    */
   public static <S, T> Transformer<S, T> constructorCall(Class<S> sourceClass, Class<T> destClass, Class<?>[] argTypes, int index, Object... args) {
      return new ConstructorCallTransformer<S, T>(sourceClass, destClass, argTypes, index, args);
   }

   private static class ConstructorCallTransformer<S, T> implements Transformer<S, T>, Serializable {
      private final Constructor<T> stconst;
      private final int            index;
      private final Object[]       args;

      public ConstructorCallTransformer(Class<S> sourceClass, Class<T> destClass, Class<?>[] argTypes, int index, Object... args) {
         this.index = index;
         this.args = args;

         try {
            stconst = destClass.getConstructor(argTypes);
         }
         catch(SecurityException e) { throw new RuntimeException(e); }
         catch(NoSuchMethodException e) { throw new RuntimeException(e); }
      }

      @Override
      public T transform(S x) {
         Throwable t;
         try {
            args[index] = x;
            return stconst.newInstance(args);
         }
         catch(IllegalArgumentException e) { t = e; }
         catch(InstantiationException e) { t = e; }
         catch(IllegalAccessException e) { t = e; }
         catch(InvocationTargetException e) { t = e; }
         throw new RuntimeException(t);
      }
   }

   /**
    * Transformer, that calls a method on the transformation object
    *
    * @param <S>        the source class
    * @param <T>        the target class
    * @param methodName the name of the method to call
    * @return a transformer
    */
   public static <S, T> Transformer<S, T> methodCall(String methodName) {
      return new MethodCallTransformer<S, T>(methodName, new Class[0], new Object[0]);
   }

   /**
    * Transformer, that calls a method on the transformation object
    *
    * @param <S>        the source class
    * @param <T>        the target class
    * @param methodName the name of the method to call
    * @param argTypes   the method argument types
    * @param args       the (constant) method arguments
    * @return a transformer
    */
   public static <S, T> Transformer<S, T> methodCall(String methodName, Class<?>[] argTypes, Object... args) {
      return new MethodCallTransformer<S, T>(methodName, argTypes, args);
   }

   private static class MethodCallTransformer<S, T> implements Transformer<S, T>, Serializable {
      private final String                      methodName;
      private final HashMap<Class<?>, Method>   methodCache;
      private final Class<?>[]                  argTypes;
      private final Object[]                    constantArgs;

      public MethodCallTransformer(String methodName, Class<?>[] argTypes, Object... constantArgs) {
         this.methodName = methodName;
         this.methodCache = new HashMap<Class<?>, Method>();
         this.argTypes = argTypes;
         this.constantArgs = constantArgs;
      }

      @SuppressWarnings("unchecked")
      @Override
      public T transform(S x) {
         if(x == null)
            return null;

         Throwable t;
         try {
            Method m = methodCache.get(x.getClass());
            if(m == null) {
               m = x.getClass().getMethod(methodName, argTypes);
               methodCache.put(x.getClass(), m);
            }
            return (T) m.invoke(x, constantArgs);
         }
         catch(SecurityException e) { t = e; }
         catch(NoSuchMethodException e) { t = e; }
         catch(IllegalArgumentException e) { t = e; }
         catch(IllegalAccessException e) { t = e; }
         catch(InvocationTargetException e) { t = e; }
         throw new RuntimeException(t);
      }
   }
   
   /**
    * Transformer, that calls a method on the given class passing the transformation
    * object as the only parameter
    * 
    * @param <S>           the source class
    * @param <T>           the target class
    * @param methodClass   the class type containing the method to call
    * @param classInstance an instance of methodClass or null for static methods
    * @param methodName    the name of the method to call
    * @param argType       the method argument type
    * @return a transformer
    */
   public static <S, T> Transformer<S, T> constMethodCall(Class<?> methodClass, Object classInstance, String methodName, Class<?> argType) {
      return new ConstMethodCallTransformer<S, T>(methodClass, classInstance, methodName, new Class<?>[] { argType }, 0, new Object[] { null });
   }

   /**
    * Transformer, that calls a method on the given class passing the transformation
    * object as parameter at the given index
    *
    * @param <S>           the source class
    * @param <T>           the target class
    * @param methodClass   the class type containing the method to call
    * @param classInstance an instance of methodClass or null for static methods
    * @param methodName    the name of the method to call
    * @param argTypes      the method argument types
    * @param index         the parameter index at which the transformation object is inserted
    * @param args          the (constant) method arguments
    * @return a transformer
    */
   public static <S, T> Transformer<S, T> constMethodCall(Class<?> methodClass, Object classInstance, String methodName, Class<?>[] argTypes, int index, Object... args) {
      return new ConstMethodCallTransformer<S, T>(methodClass, classInstance, methodName, argTypes, index, args);
   }

   private static class ConstMethodCallTransformer<S, T> implements Transformer<S, T>, Serializable {
      private final Object   classInstance;
      private final int      index;
      private final Object[] args;
      private final Method   method;

      public ConstMethodCallTransformer(Class<?> methodClass, Object classInstance, String methodName, Class<?>[] argTypes, int index, Object... args) {
         this.classInstance = classInstance;
         this.index = index;
         this.args = args;

         try {
            this.method = methodClass.getMethod(methodName, argTypes);
         }
         catch(SecurityException e) { throw new RuntimeException(e); }
         catch(NoSuchMethodException e) { throw new RuntimeException(e); }
      }

      @SuppressWarnings("unchecked")
      @Override
      public T transform(S x) {
         if(x == null)
            return null;

         Throwable t;
         try {
            args[index] = x;
            return (T) method.invoke(classInstance, args);
         }
         catch(IllegalArgumentException e) {t = e;}
         catch(IllegalAccessException e) {t = e;}
         catch(InvocationTargetException e) {t = e;}
         throw new RuntimeException(t);
      }
   }

   /**
    * Create a transformer, which performs an EL evaluation on its input object,
    * and returns the result.
    * 
    * @param <S>            the source class
    * @param <T>            the target class
    * @param tClass         the target class
    * @param accessName     the input object's name in the EL-expression, e.g.
    *                       "theobject", if the expression is like "${theobject.aproperty}"
    * @param expression     the expression to evaluate
    * @return a transformer
    */
   public static <S, T> Transformer<S, T> el(Class<T> tClass, String accessName, String expression) {
      return new ELTransformer<S, T>(tClass, accessName, expression);
   }

   private static class ELTransformer<S, T> implements Transformer<S, T>, Serializable {
      private Class<T>          tClass;
      private String            accessName;
      private String            expression;
      private ELAccessor        elAccessor;

      private ELTransformer(Class<T> tClass, String accessName, String expression) {
         this.tClass     = tClass;
         this.accessName = accessName;
         this.expression = expression;

         elAccessor = new ELAccessor();
      }

      @Override
      public T transform(S s) {
         elAccessor.setVariable(accessName, s);
         return elAccessor.evaluateExpression(expression, tClass);
      }
   }
   
   /**
    * Create a transformer, which initializes a new bean object of type targetClass with properties
    * read from the transformation source object and some constant properties. The new object is created
    * using its default constructor and initialized via standard bean property setters.
    * 
    * @param <S>            the source class
    * @param <T>            the target class
    * @param targetClass    the target class
    * @param propMap        a map consisting of target property to source
    *                       property assignments or null (i.e. the key is the
    *                       target property, the value is the source property)
    * @param constMap       a map containing all constant property values or null
    * @return a transformer
    */
   public static <S, T> Transformer<S, T> beanInit(Class<T> targetClass, Map<String, String> propMap, Map<String, Object> constMap) {
      return new BeanInitTransformer<S, T>(targetClass, propMap, constMap);
   }
   
   private static class BeanInitTransformer<S, T> implements Transformer<S, T>, Serializable {
      private Class<T>  targetClass;
      private Map<String, String> propMap;
      private Map<String, Object> constMap;
      
      private BeanInitTransformer(Class<T> targetClass, Map<String, String> propMap, Map<String, Object> constMap) {
         this.targetClass = targetClass;
         this.propMap = propMap;
         this.constMap = constMap;
      }

      @Override
      public T transform(S s) {
         Throwable x;
         try {
            T t = targetClass.newInstance();
            if(constMap != null)
               for(String prop : constMap.keySet())
                  PropertyUtils.setProperty(t, prop, constMap.get(prop));
            if(propMap != null)
               for(String prop : propMap.keySet())
                  PropertyUtils.setProperty(t, prop, PropertyUtils.getProperty(s, propMap.get(prop)));
            return t;
         }
         catch(InstantiationException e) { x = e; }
         catch(IllegalAccessException e) { x = e; }
         catch(InvocationTargetException e) { x = e; }
         catch(NoSuchMethodException e) { x = e; }
         throw new RuntimeException(x);
      }
   }
   
   
   private static final class PairedTransformer<A, X, B, Y> implements Transformer<Pair<A, X>, Pair<B, Y>>, Serializable {
      private final Transformer<X, Y> ty;
      private final Transformer<A, B> tx;
   
      private PairedTransformer(Transformer<X, Y> ty, Transformer<A, B> tx) {
         this.ty = ty;
         this.tx = tx;
      }
   
      @Override
      public Pair<B, Y> transform(Pair<A, X> s) {
         return new Pair<B, Y>(tx.transform(s.x), ty.transform(s.y));
      }
   }

   /**
    * Create a transformer for the two components of a pair using two given
    * transformers for Pair.x and Pair.y
    * 
    * @param <A> source pair first type
    * @param <X> source pair second type
    * @param <B> destination pair first type
    * @param <Y> destination pair second type
    * @param tx first transformer
    * @param ty second transformer
    * @return a transformer, that returns (tx(x), ty(y)) for an input pair (x, y)
    */
   public static <A, B, X, Y> Transformer<Pair<A, X>, Pair<B, Y>> paired(final Transformer<A, B> tx, final Transformer<X, Y> ty) {
      return new PairedTransformer<A, X, B, Y>(ty, tx);
   }
   
   
   private static final class ChainLinkTransformer<A, C, B> implements Transformer<A, C>, Serializable {
      private final Transformer<A, B> t1;
      private final Transformer<B, C> t2;
   
      private ChainLinkTransformer(Transformer<A, B> t1, Transformer<B, C> t2) {
         this.t1 = t1;
         this.t2 = t2;
      }
   
      @Override
      public C transform(A a) {
         return t2.transform(t1.transform(a));
      }
   }

   /**
    * Create a transformer, which combines the transformations of two given
    * transformers.
    * 
    * @param <A> source class
    * @param <B> intermediate class
    * @param <C> result class
    * @param t1 a transformer from A to B
    * @param t2 a transformer from B to C
    * @return (t1 * t2)(x)
    */
   public static <A, B, C> Transformer<A, C> chainLink(final Transformer<A, B> t1, final Transformer<B, C> t2) {
      return new ChainLinkTransformer<A, C, B>(t1, t2);
   }
   
   private static class MapLookupTransformer<K, V> implements Transformer<K, V>, Serializable {
      private Map<K, V> backingMap;

      private MapLookupTransformer(Map<K, V> backingMap) {
         this.backingMap = backingMap;
      }

      @Override
      public V transform(K s) {
         return backingMap.get(s);
      }
   }
   
   
   /**
    * Create a transformer, which performs a lookup on a given map to determine
    * the transformation result
    *
    * @param <K>          transformer source == map key class
    * @param <V>          transformer target == map value class
    * @param backingMap   the map
    * @return a transformer, which calls <code>get</code> on the backing map
    */
   public static <K, V> Transformer<K, V> mapLookup(Map<K, V> backingMap) {
      return new MapLookupTransformer<K, V>(backingMap);
   }
   
   private static class ToPairTransformer<A, X, Y> implements Transformer<A, Pair<X, Y>>, Serializable {
      private Transformer<A, X>   xTransformer;
      private Transformer<A, Y>   yTransformer;
      
      private ToPairTransformer(Transformer<A, X> xTransformer, Transformer<A, Y> yTransformer) {
         this.xTransformer = xTransformer;
         this.yTransformer = yTransformer;
      }

      @Override
      public Pair<X, Y> transform(A s) {
         return new Pair<X, Y>(xTransformer.transform(s), yTransformer.transform(s));
      }
   }

   /**
    * Create a transformer, which uses two component-transformers to build a
    * pair of the two result objects
    * 
    * @param <A>  the source class
    * @param <X>  destination class 1
    * @param <Y>  destination class 2
    * @param xTransformer  component transformer 1
    * @param yTransformer  component transformer 2
    * @return a transformer, which generates a pair from xTransformer.transform
    * and yTransformer.transform
    */
   public static <A, X, Y> Transformer<A, Pair<X, Y>> toPair(Transformer<A, X> xTransformer, Transformer<A, Y> yTransformer) {
      return new ToPairTransformer<A, X, Y>(xTransformer, yTransformer);
   }
}
